from django.core.management import BaseCommand
from django.db import DEFAULT_DB_ALIAS

from .. import helpers


class Command(BaseCommand):
    help = """
    Создаёт отсутствующие файлы вариаций.
    """
    requires_system_checks = []

    def add_arguments(self, parser):
        parser.add_argument(
            "--database",
            action="store",
            dest="database",
            default=DEFAULT_DB_ALIAS,
            help="Nominates the database to use. Defaults to the 'default' database.",
        )
        parser.add_argument(
            "--async",
            action="store_true",
            help="Use django-rq to create variation files",
        )

    def handle(self, *args, **options):
        helpers.create_missing_variations(
            async_=options["async"],
            database=options["database"]
        )



import sys
from typing import Type, cast

from anytree import LevelOrderIter
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import FieldDoesNotExist
from django.core.management import BaseCommand
from django.db import DEFAULT_DB_ALIAS

from ... import exceptions, helpers
from ...models.base import FileFieldResource, FileResource, VersatileImageResourceMixin
from ...models.collection import Collection, CollectionItemBase
from ...models.mixins import BacklinkModelMixin


class Command(BaseCommand):
    help = """
    Проверка целостности экземпляров файловых моделей.
    """
    options = None
    verbosity = None
    database = DEFAULT_DB_ALIAS

    def add_arguments(self, parser):
        parser.add_argument(
            "--database",
            action="store",
            dest="database",
            default=DEFAULT_DB_ALIAS,
            help="Nominates the database to use. Defaults to the 'default' database.",
        )
        parser.add_argument(
            "-o", "--check-ownership",
            action="store_true",
            default=False,
            help="Check `owner_XXX` fields.",
        )
        parser.add_argument(
            "-f", "--check-files-exist",
            action="store_true",
            default=False,
            help="Check file existence.",
        )
        parser.add_argument(
            "-c", "--check-content-types",
            action="store_true",
            default=False,
            help="Check collection ContentType.",
        )
        parser.add_argument(
            "-t", "--check-item-types",
            action="store_true",
            default=False,
            help="Check item `type` values.",
        )

    def handle(self, *args, **options):
        self.options = options
        self.verbosity = options["verbosity"]
        self.database = options["database"]
        self.run_steps()

    def run_steps(self):
        check_ownership = self.options["check_ownership"]
        check_files_exist = self.options["check_files_exist"]
        check_content_types = self.options["check_content_types"]
        check_item_types = self.options["check_item_types"]

        check_all = not any([
            check_ownership,
            check_files_exist,
            check_content_types,
            check_item_types,
        ])

        if check_all or check_ownership:
            self.check_ownership()

        if check_all or check_files_exist:
            self.check_files_exist()

        if check_all or check_content_types:
            self.check_content_types()

        if check_all or check_item_types:
            self.check_item_types()

    def check_ownership(self):
        """
        Проверяет, что ресурсы и коллекции
        *) имеют значения owner_app_label и owner_model_name, ссылающиеся на существующую модель
        *) имя поля, указанное в owner_fieldname существует в классе модели владельца
        """
        for root in helpers.get_resource_model_trees():
            for node in reversed(tuple(LevelOrderIter(root))):
                model = node.model
                self._check_model_ownership(model)

        for root in helpers.get_collection_trees(include_proxy=True):
            for node in reversed(tuple(LevelOrderIter(root))):
                model = node.model

                if model is Collection:
                    continue

                self._check_model_ownership(model)

    def _check_model_ownership(self, model: Type[BacklinkModelMixin]):
        if not issubclass(model, BacklinkModelMixin):
            return

        queryset = model.objects.using(self.database).only(
            "pk",
            "owner_app_label",
            "owner_model_name",
            "owner_fieldname"
        )
        for instance in queryset.iterator():
            owner_model = instance.get_owner_model()
            if owner_model is None:
                print(
                    "\033[31mERROR\033[0m: "
                    "\033[92m{}.{}\033[0m #{} has an invalid owner reference: "
                    "\033[91m{}.{}\033[0m".format(
                        type(instance)._meta.app_label,
                        type(instance).__name__,
                        instance.pk,
                        instance.owner_app_label,
                        instance.owner_model_name,
                    ),
                    file=sys.stderr
                )
                sys.stderr.flush()
                continue

            if not instance.owner_fieldname:
                print(
                    "\033[31mERROR\033[0m: "
                    "\033[92m{}.{}\033[0m #{} has an empty \033[92mowner_fieldname\033[0m field".format(
                        type(instance)._meta.app_label,
                        type(instance).__name__,
                        instance.pk,
                    ),
                    file=sys.stderr
                )
                sys.stderr.flush()
                continue

            try:
                owner_model._meta.get_field(instance.owner_fieldname)
            except FieldDoesNotExist:
                print(
                    "\033[31mERROR\033[0m: "
                    "\033[92m{}.{}\033[0m #{} references a field name that does not exist: \033[91m{}\033[0m".format(
                        type(instance)._meta.app_label,
                        type(instance).__name__,
                        instance.pk,
                        instance.owner_fieldname
                    ),
                    file=sys.stderr
                )
                sys.stderr.flush()
                continue

    def check_content_types(self):
        """
        Проверяет, что для коллекций
        *) существует модель, заданная через ContentType.
        """
        for root in helpers.get_collection_trees():
            for node in reversed(tuple(LevelOrderIter(root))):
                model = node.model
                self._check_content_types(model)

    def _check_content_types(self, model: Type[Collection]):
        queryset = model.objects.using(self.database)
        for instance in queryset.iterator():
            content_type_id = instance.collection_content_type_id
            content_type = ContentType.objects.get_for_id(content_type_id)
            model_class = content_type.model_class()
            if model_class is None:
                print(
                    "\033[31mERROR\033[0m: "
                    "Collection #{} refers to a model \033[92m{}.{}\033[0m that does not exist.".format(
                        instance.pk,
                        content_type.app_label,
                        content_type.model
                    ),
                    file=sys.stderr
                )
                sys.stderr.flush()

    def check_files_exist(self):
        """
        Проверяет, что экземпляры файловых ресурсов (UploadedFile, UploadedImage и т.п.)
        и файловых элементов коллекций ссылаются на существующие файлы.

        P.S.: Не проверяет существование вариаций изображений!
        """
        for root in helpers.get_resource_model_trees():
            for node in reversed(tuple(LevelOrderIter(root))):
                model = node.model
                self._check_files_exist(model)

    def _check_files_exist(self, model: Type[FileResource]):
        if not issubclass(model, FileResource):
            return

        queryset = model.objects.using(self.database)
        query_fields = []

        if issubclass(model, FileFieldResource):
            model = cast(Type[FileFieldResource], model)
            file_field = model.get_file_field()
            query_fields.append(file_field.name)

        if issubclass(model, VersatileImageResourceMixin):
            # Для инициализации экземпляров VersatileImageResourceMixin нужны
            # дополнительные поля - для создания полей вариаций.
            if issubclass(model, BacklinkModelMixin):
                query_fields.extend([
                    "owner_app_label",
                    "owner_model_name",
                    "owner_fieldname"
                ])

            if issubclass(model, CollectionItemBase):
                query_fields.extend([
                    "collection_content_type_id",
                    "polymorphic_ctype_id",
                    "type",
                ])

        queryset = queryset.only(*query_fields)

        for instance in queryset.iterator():
            if not instance.file_exists():
                print(
                    "\033[31mERROR\033[0m: "
                    "\033[92m{}.{}\033[0m #{} references a file that does not exist".format(
                        type(instance)._meta.app_label,
                        type(instance).__name__,
                        instance.pk
                    ),
                    file=sys.stderr
                )
                sys.stderr.flush()

    def check_item_types(self):
        """
        Проверяет, что элементы коллекций
        *) имеют значение type, которое присутствует в коллекции
        *) имеют класс, соответствующий модели, указанной для данного type
        """
        queryset = CollectionItemBase.objects.using(self.database)
        for item in queryset.iterator():
            try:
                collection_cls = item.get_collection_class()
            except exceptions.CollectionModelNotFoundError:
                continue

            if item.type not in collection_cls.item_types:
                print(
                    "\033[31mERROR\033[0m: "
                    "\033[92m{}.{}\033[0m #{} has a type value that has not been declared "
                    "in the \033[92m{}.{}\033[0m: \033[91m{}\033[0m".format(
                        type(item)._meta.app_label,
                        type(item).__name__,
                        item.pk,
                        collection_cls._meta.app_label,
                        collection_cls.__name__,
                        item.type,
                    ),
                    file=sys.stderr
                )
                sys.stderr.flush()
                continue

            item_model = collection_cls.get_item_model(item.type)
            if not issubclass(type(item), item_model):
                print(
                    "\033[31mERROR\033[0m: "
                    "\033[92m{}.{}\033[0m #{} has a class type different from that "
                    "specified in the \033[92m{}.{}\033[0m as \033[92m{}\033[0m type: "
                    "\033[91m{}.{}\033[0m".format(
                        type(item)._meta.app_label,
                        type(item).__name__,
                        item.pk,
                        collection_cls._meta.app_label,
                        collection_cls.__name__,
                        item.type,
                        item_model._meta.app_label,
                        item_model.__name__,
                    ),
                    file=sys.stderr
                )
                sys.stderr.flush()
                continue



# Generated by Django 4.0.2 on 2022-02-09 05:38

from django.db import migrations, models
import django.db.models.deletion
import paper_uploads.models.fields.base


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('paper_uploads', '0008_rename_item_type_collectionitembase_type'),
    ]

    operations = [
        migrations.AlterField(
            model_name='collectionitembase',
            name='polymorphic_ctype',
            field=models.ForeignKey(editable=False, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='polymorphic_%(app_label)s.%(class)s_set+', to='contenttypes.contenttype'),
        ),
        migrations.AlterField(
            model_name='fileitem',
            name='file',
            field=paper_uploads.models.fields.base.DynamicStorageFileField(verbose_name='file'),
        ),
        migrations.AlterField(
            model_name='svgitem',
            name='file',
            field=paper_uploads.models.fields.base.DynamicStorageFileField(verbose_name='file'),
        ),
        migrations.AlterField(
            model_name='uploadedfile',
            name='file',
            field=paper_uploads.models.fields.base.DynamicStorageFileField(verbose_name='file'),
        ),
    ]



import os
import shutil
import tempfile
from typing import Any, Dict, Iterable, List, Optional, Union
from uuid import UUID

from django.conf import settings
from django.core.exceptions import (
    NON_FIELD_ERRORS,
    ImproperlyConfigured,
    MultipleObjectsReturned,
    ObjectDoesNotExist,
    ValidationError,
)
from django.core.files.uploadedfile import UploadedFile
from django.core.handlers.wsgi import WSGIRequest
from django.http import HttpResponse, JsonResponse
from django.template import loader
from django.utils.decorators import method_decorator
from django.utils.functional import Promise
from django.utils.translation import gettext_lazy as _
from django.views import View
from django.views.decorators.csrf import csrf_exempt
from django.views.generic.edit import FormMixin

from .. import exceptions
from ..files import TemporaryUploadedFile
from ..logging import logger
from ..models.base import Resource


class AjaxView(View):
    @staticmethod
    def success_response(data: Optional[Dict[str, Any]] = None) -> JsonResponse:
        data = data or {}
        return JsonResponse(data)

    @staticmethod
    def get_exception_messages(exception: ValidationError) -> List[str]:
        messages = []  # type: List[str]
        for msg in exception:
            if isinstance(msg, tuple):
                field, errors = msg
                if field == NON_FIELD_ERRORS:
                    for error in reversed(errors):
                        messages.insert(0, error)
                else:
                    messages.extend("'{}': {}".format(field, error) for error in errors)
            else:
                messages.append(msg)
        return messages

    @staticmethod
    def error_response(
        errors: Union[str, Iterable[str]] = "", **extra_data
    ) -> JsonResponse:
        if not errors:
            errors = []
        elif isinstance(errors, str):
            errors = [errors]
        elif isinstance(errors, Promise):
            errors = [str(errors)]

        data = {
            "errors": errors,
        }
        data.update(extra_data)
        return JsonResponse(data)

    @classmethod
    def wrap(cls, func):
        def inner(*args, **kwargs) -> HttpResponse:
            try:
                return func(*args, **kwargs)
            except exceptions.InvalidContentType as e:
                logger.exception("Error")
                return cls.error_response(_("Invalid ContentType: %s") % e.value)
            except exceptions.InvalidObjectId as e:
                logger.exception("Error")
                return cls.error_response(_("Invalid ID: %s") % e.value)
            except exceptions.InvalidItemType as e:
                logger.exception("Error")
                return cls.error_response(_("Invalid itemType: %s") % e.value)
            except ObjectDoesNotExist:
                logger.exception("Error")
                return cls.error_response(_("Object not found"))
            except MultipleObjectsReturned:
                logger.exception("Error")
                return cls.error_response(_("Multiple objects returned"))
            except ValidationError as e:
                messages = cls.get_exception_messages(e)
                logger.debug(messages)
                return cls.error_response(messages)
            except Exception as e:
                logger.exception("Error")
                if hasattr(e, "args"):
                    message = "{}: {}".format(type(e).__name__, e.args[0])
                else:
                    message = type(e).__name__
                return cls.error_response(message)

        return inner


class UploadFileViewBase(AjaxView):
    http_method_names = ["post"]

    @method_decorator(csrf_exempt)
    def dispatch(self, *args, **kwargs):
        return super().dispatch(*args, **kwargs)

    def post(self, request: WSGIRequest, *args, **kwargs) -> HttpResponse:
        if not request.user.has_perm("paper_uploads.upload"):
            return self.error_response(_("Access denied"))

        try:
            file = self.upload_chunk(request)
        except exceptions.ContinueUpload:
            return self.success_response()
        except exceptions.UncompleteUpload:
            return self.error_response(preventRetry=True)
        except exceptions.InvalidUUID as e:
            logger.exception("Error")
            return self.error_response(_("Invalid UUID: %s") % e.value, preventRetry=True)
        except exceptions.InvalidChunking:
            logger.exception("Error")
            return self.error_response(_("Invalid chunking"), preventRetry=True)

        try:
            return self.wrap(self.handle)(file)
        except exceptions.UnsupportedResource as e:
            return self.error_response(e.message)
        finally:
            file.close()

    def upload_chunk(self, request: WSGIRequest) -> UploadedFile:
        try:
            chunk_index = int(request.POST["paperChunkIndex"])
            total_chunks = int(request.POST["paperTotalChunkCount"])
        except KeyError:
            # small file, no chunks
            chunk_index = 0
            total_chunks = 1
        except (ValueError, TypeError):
            raise exceptions.InvalidChunking

        uuid = request.POST.get("paperUUID")
        try:
            uid = UUID(uuid)
        except (AttributeError, ValueError):
            raise exceptions.InvalidUUID(uuid)

        tempdir = request.session.get("paper_uploads_tempdir")
        if tempdir is None or not os.path.isdir(tempdir):
            if request.user.pk is not None:
                tempdir_suffix = ".user_{}".format(request.user.pk)
            else:
                tempdir_suffix = None

            tempdir = tempfile.mkdtemp(
                prefix="paper_uploads.",
                suffix=tempdir_suffix,
                dir=settings.FILE_UPLOAD_TEMP_DIR
            )
            request.session["paper_uploads_tempdir"] = tempdir

        tempfilepath = os.path.join(tempdir, str(uid))
        file = request.FILES.get("file")
        if file is None:
            # случается при отмене загрузки на медленном интернете
            if os.path.isfile(tempfilepath):
                os.unlink(tempfilepath)
            raise exceptions.UncompleteUpload

        if total_chunks > 1:
            with open(tempfilepath, "a+b") as fp:
                shutil.copyfileobj(file, fp)

            if chunk_index < total_chunks - 1:
                raise exceptions.ContinueUpload

            file = TemporaryUploadedFile(
                open(tempfilepath, "rb"),
                name=os.path.basename(file.name),
                size=os.path.getsize(tempfilepath)
            )
        return file

    def handle(self, file: UploadedFile) -> HttpResponse:
        raise NotImplementedError


class DeleteFileViewBase(AjaxView):
    http_method_names = ["post"]

    @method_decorator(csrf_exempt)
    def dispatch(self, *args, **kwargs):
        return super().dispatch(*args, **kwargs)

    def post(self, request: WSGIRequest, *args, **kwargs) -> HttpResponse:
        if not request.user.has_perm("paper_uploads.delete"):
            return self.error_response(_("Access denied"))
        return self.wrap(self.handle)()

    def handle(self) -> HttpResponse:
        raise NotImplementedError


class ChangeFileViewBase(FormMixin, AjaxView):
    template_name = None
    http_method_names = ["get", "post"]

    def get_template_name(self) -> str:
        if self.template_name is None:
            raise ImproperlyConfigured(
                "{} requires either a definition of "
                "'template_name' or an implementation of 'get_template_name()'".format(
                    type(self).__name__
                )
            )
        else:
            return self.template_name

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context.update({
            "action": self.request.get_full_path(),
        })
        return context

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs["instance"] = self.get_instance()
        return kwargs

    def get_instance(self):
        raise NotImplementedError

    def get(self, request: WSGIRequest, *args, **kwargs) -> HttpResponse:
        if not request.user.has_perm("paper_uploads.change"):
            return self.error_response(_("Access denied"))

        form_response = self.wrap(self.render_form)()
        if isinstance(form_response, HttpResponse):
            return form_response

        return self.success_response({
            "form": form_response
        })

    def render_form(self, **kwargs) -> str:
        context = self.get_context_data(**kwargs)
        template_name = self.get_template_name()
        return loader.render_to_string(template_name, context, request=self.request)

    def post(self, request: WSGIRequest, *args, **kwargs) -> HttpResponse:
        if not request.user.has_perm("paper_uploads.change"):
            return self.error_response(_("Access denied"))

        return self.wrap(self.validate_form)()

    def validate_form(self) -> HttpResponse:
        form = self.get_form()
        if not form.is_valid():
            return self.form_invalid(form)

        return self.form_valid(form)

    def form_valid(self, form) -> HttpResponse:
        try:
            instance = form.save()
        except FileNotFoundError as e:
            error = _("File not found: %s") % e.args[0] if e.args else "???"
            logger.debug(error)
            return self.error_response(error)

        return self.success(instance)

    def form_invalid(self, form) -> HttpResponse:
        return JsonResponse({
            "form_errors": form.errors.get_json_data()
        })

    def success(self, instance: Resource) -> HttpResponse:
        return self.success_response(instance.as_dict())



# Generated by Django 4.0.1 on 2022-02-17 14:09

from django.db import migrations, models
import django.utils.timezone


class Migration(migrations.Migration):

    dependencies = [
        ('paper_uploads', '0003_alter_fileitem_resource_name_and_more'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='fileitem',
            name='created_at',
        ),
        migrations.RemoveField(
            model_name='fileitem',
            name='modified_at',
        ),
        migrations.RemoveField(
            model_name='imageitem',
            name='created_at',
        ),
        migrations.RemoveField(
            model_name='imageitem',
            name='modified_at',
        ),
        migrations.RemoveField(
            model_name='svgitem',
            name='created_at',
        ),
        migrations.RemoveField(
            model_name='svgitem',
            name='modified_at',
        ),
        migrations.AddField(
            model_name='collectionitembase',
            name='created_at',
            field=models.DateTimeField(default=django.utils.timezone.now, editable=False, verbose_name='created at'),
        ),
        migrations.AddField(
            model_name='collectionitembase',
            name='modified_at',
            field=models.DateTimeField(auto_now=True, verbose_name='changed at'),
        ),
    ]



import warnings


class CollectionModelNotFoundError(Exception):
    """
    Исключение, вызываемое методом `get_collection_class()` элемента коллекции,
    когда ContentType коллекции существует, но модель коллекции удалена.
    """
    pass


class CollectionItemNotFoundError(Exception):
    """
    Исключение, вызываемое методом `get_item_type_field()` элемента коллекции,
    когда к классе коллекции не нашлось соответствующего поля для текущего элемента.
    """
    pass


class UnsupportedCollectionItemError(Exception):
    """
    Исключение, вызываемое методом `attach_to()` элемента коллекции,
    когда класс текущего элемента не подходит ни к одному из возможных
    типов коллекции.
    """
    pass


class InvalidParameter(Exception):
    def __init__(self, value):
        self.value = value
        super().__init__()


class InvalidContentType(InvalidParameter):
    pass


class InvalidItemType(InvalidParameter):
    pass


class InvalidObjectId(InvalidParameter):
    pass


class ContinueUpload(Exception):
    pass


class UncompleteUpload(Exception):
    pass


class InvalidUUID(Exception):
    def __init__(self, value):
        self.value = value
        super().__init__()


class InvalidChunking(Exception):
    pass


class UnsupportedResource(Exception):
    def __init__(self, message):
        self.message = message
        super().__init__(message)


class UnsupportedFileError(UnsupportedResource):
    def __init__(self, *args, **kwargs):
        warnings.warn(
            "UnsupportedFileError is deprecated in favor of UnsupportedResource",
            DeprecationWarning,
            stacklevel=2
        )
        super().__init__(*args, **kwargs)



import operator
import sys
from datetime import timedelta
from itertools import groupby
from collections import namedtuple
from typing import Any, Callable, Generator, List, Tuple, Type, Union

from anytree import LevelOrderIter
from django.apps import apps
from django.contrib.contenttypes.models import ContentType
from django.db import DEFAULT_DB_ALIAS, models
from django.utils.timezone import now

from .. import helpers
from ..models.base import Resource
from ..models.collection import Collection, CollectionBase, CollectionItemBase
from . import utils
from .prompt import prompt_action, prompt_variants


ModelChoice = namedtuple("ModelChoice", ["name", "type"])
ItemTypeChoice = namedtuple("ItemTypeChoice", ["name", "field"])
FieldChoice = namedtuple("FieldChoice", ["name", "field"])
VariationChoice = namedtuple("VariationChoice", ["name", "size"])


def find_empty_collections(
    min_age: int = 24 * 3600,
    database: str = DEFAULT_DB_ALIAS,
) -> Generator[models.QuerySet[Collection], Any, None]:
    """
    Поиск коллекций, которые не содержат элементов.

    Для того, чтобы не удалить коллекции, которые созданы недавно и ещё не
    заполнены, можно воспользоваться параметром `min_age`. Он задаёт минимальный
    возраст коллекции в секундах.

    :param min_age: минимальное время с момента создания коллекции в секундах.
    :param database: алиас базы данных для поиска коллекций.
    """
    for root in helpers.get_collection_trees():
        for node in reversed(tuple(LevelOrderIter(root))):
            model = node.model
            concrete_ct = ContentType.objects.db_manager(using=database).get_for_model(model)

            created_before = now() - timedelta(seconds=min_age)
            queryset = model.objects.using(database).filter(
                ~models.Exists(CollectionItemBase.objects.using(database).filter(
                    concrete_collection_content_type=concrete_ct,
                    collection_id=models.OuterRef("pk")
                )),
                concrete_collection_content_type=concrete_ct,
                created_at__lte=created_before
            ).select_related(
                "collection_content_type"
            ).only(
                "id", "collection_content_type"
            ).order_by("collection_content_type_id")

            if queryset.exists():
                for ct, group in groupby(queryset.iterator(), key=lambda x: x.collection_content_type):
                    collection_cls = ct.model_class()
                    ids = set(item.pk for item in group)
                    yield collection_cls.objects.using(database).filter(pk__in=ids)


def remove_empty_collections(
    min_age: int = 24 * 3600,
    database: str = DEFAULT_DB_ALIAS
):
    """
    Удаление коллекций, которые не содержат элементов.

    Удалять недавно созданные коллекции нельзя, т.к. они могут быть в процессе
    заполнения другим пользователем.

    :param min_age: минимальное время с момента создания коллекции в секундах.
    :param database: алиас базы данных для поиска коллекций.
    """
    for collection_qs in find_empty_collections(min_age=min_age, database=database):
        collection_cls = collection_qs.model
        count = collection_qs.count()

        while True:
            action = prompt_action(
                message=(
                    "Found \033[92m%(count)d %(app_label)s.%(model)s\033[0m collection "
                    "that do not have any elements.\n"
                    "What would you like to do with it?"
                    if count == 1 else
                    "Found \033[92m%(count)d %(app_label)s.%(model)s\033[0m collections "
                    "that do not have any elements.\n"
                    "What would you like to do with them?"
                ) % {
                    "count": count,
                    "app_label": collection_cls._meta.app_label,
                    "model": collection_cls.__name__,
                },
                choices=["Skip", "Print", "Delete", "Exit"]
            )

            if action == "Print":
                print("-" * 48)
                for index, obj in enumerate(collection_qs.iterator(), start=1):
                    number = "{})".format(index)
                    print(
                        "{:<3} \033[92m{}.{}\033[0m #{}\n"
                        "    Created at: {}".format(
                            number,
                            obj._meta.app_label,
                            obj.__class__.__name__,
                            obj.pk,
                            obj.created_at.strftime("%Y-%m-%d %H:%M:%S"),
                        )
                    )
                print("-" * 48)
            elif action == "Delete":
                collection_qs.delete()
                break
            elif action == "Skip":
                break
            else:
                return


def find_missing_variations(
    database: str = DEFAULT_DB_ALIAS
) -> Generator[Tuple[Resource, List[str]], Any, None]:
    """
    Поиск экземпляров изображений, для которых отсутствует хотя бы один
    файл вариации.

    :param database: алиас базы данных для поиска коллекций.
    """
    for root in helpers.get_resource_model_trees():
        for node in reversed(tuple(LevelOrderIter(root))):
            model = node.model

            if not utils.is_variations_allowed(model):
                continue

            for instance in model.objects.using(database).iterator():
                missing_variations = []
                for vname, vfile in instance.variation_files():
                    if not vfile.exists():
                        missing_variations.append(vname)

                if missing_variations and instance.file_exists():
                    yield instance, missing_variations


def create_missing_variations(
    async_: bool = False,
    database: str = DEFAULT_DB_ALIAS
):
    """
    Создание отсутствующих файлов вариаций.

    :param async_: использовать django-rq для создания файлов вариаций.
    :param database: алиас базы данных для поиска файлов.
    """
    for instance, missing_variations in find_missing_variations(database):
        count = len(missing_variations)

        print(
            "Creating \033[92m%(count)d\033[0m %(label)s"
            " for \033[92m%(app_label)s.%(model)s\033[0m #%(pk)s ... " % {
                "count": count,
                "label": "variation" if count == 1 else "variations",
                "app_label": type(instance)._meta.app_label,
                "model": type(instance).__name__,
                "pk": instance.pk
            },
            end=""
        )
        sys.stdout.flush()

        if async_:
            instance.recut_async(names=missing_variations)
        else:
            instance.recut(names=missing_variations)

        print("done")
        sys.stdout.flush()


def select_resource_model(
    message: str = None,
    multiple: bool = False,
    predicate: Callable = None,
    prepend_choices: List[str] = None,
    append_choices: List[str] = None,
    default: Union[str, List[str]] = None
) -> Union[None, str, List[str]]:
    """
    Диалог выбора одной или нескольких моделей, связанных с файловыми ресурсами.

    В списке выводятся регулярные модели, содержащие хотя бы одно поле,
    ссылающееся на подкласс Resource, а также модели коллекций.

    Пример:
        from paper_uploads.management.helpers import *
        from paper_uploads.management import utils

        model = select_resource_model()
        if utils.is_collection(model):
            item_type = select_collection_item_type(model)
            variation_name = select_collection_variations(model, item_type)
        else:
            field_name = select_resource_field(model)
            variation_name = select_resource_variations(model, field_name)
    """
    choices_data = []
    for model in apps.get_models():
        opts = model._meta

        if predicate is not None and not predicate(model):
            continue

        for field in opts.get_fields(include_hidden=True):
            if not utils.is_resource_field(field):
                continue

            choices_data.append(
                ModelChoice(
                    name="{}.{}".format(opts.app_label, opts.model_name),
                    type="Regular",
                )
            )
            break

    collection_models = (
        node.model
        for root in helpers.get_collection_trees(include_proxy=True)
        for node in reversed(tuple(LevelOrderIter(root)))
    )
    for model in collection_models:
        opts = model._meta

        if predicate is not None and not predicate(model):
            continue

        choices_data.append(
            ModelChoice(
                name="{}.{}".format(opts.app_label, opts.model_name),
                type="Collection",
            )
        )

    choices_data = sorted(
        choices_data,
        key=operator.attrgetter("name")
    )

    # Оформление опций в виде двух колонок.
    rows = []
    max_model_length = max((len(record.name) for record in choices_data), default=0)
    max_model_length = min(max(max_model_length, 10), 48)
    for choice in choices_data:
        rows.append([
            "{:<{width}} {}".format(
                choice.name,
                choice.type,
                width=max_model_length + 1
            ),
            choice.name
        ])

    if prepend_choices:
        rows = list(prepend_choices) + rows

    if append_choices:
        rows.extend(append_choices)

    if not rows:
        return

    if multiple:
        return prompt_variants(
            message=message or (
                "Select one or more models:\n"
                "   {column_name:<{width}} Type\n"
                "   {line}".format(
                    column_name="Model",
                    line="-" * (max_model_length + 16),
                    width=max_model_length + 1
                )
            ),
            choices=rows,
            default=default
        )
    else:
        return prompt_action(
            message=message or (
                "Select model:\n"
                "   {column_name:<{width}} Type\n"
                "   {line}".format(
                    column_name="Model",
                    line="-" * (max_model_length + 16),
                    width=max_model_length + 1
                )
            ),
            choices=rows,
            default=default
        )


def select_collection_item_type(
    collection: Type[CollectionBase],
    message: str = None,
    multiple: bool = False,
    predicate: Callable = None,
    prepend_choices: List[str] = None,
    append_choices: List[str] = None,
    default: Union[str, List[str]] = None
) -> Union[None, str, List[str]]:
    """
    Диалог выбора одного или нескольких типов элементов коллекции.
    """
    choices_data = []
    for name, field in collection.item_types.items():
        if predicate is not None and not predicate(field):
            continue

        choices_data.append(
            ItemTypeChoice(
                name=name,
                field=field
            )
        )

    choices_data = sorted(
        choices_data,
        key=operator.attrgetter("name")
    )

    # Оформление опций в виде двух колонок.
    rows = []
    max_name_length = max((len(record.name) for record in choices_data), default=0)
    max_name_length = min(max(max_name_length, 10), 48)
    for choice in choices_data:
        rows.append([
            "{:<{width}} {}".format(
                choice.name,
                choice.field.model.__name__,
                width=max_name_length + 1
            ),
            choice.name
        ])

    if prepend_choices:
        rows = list(prepend_choices) + rows

    if append_choices:
        rows.extend(append_choices)

    if not rows:
        return

    if multiple:
        return prompt_variants(
            message=message or (
                "Select one or more item types:\n"
                "   {column_name:<{width}} Model\n"
                "   {line}".format(
                    column_name="Name",
                    line="-" * (max_name_length + 16),
                    width=max_name_length + 1
                )
            ),
            choices=rows,
            default=default
        )
    else:
        return prompt_action(
            message=message or (
                "Select item type:\n"
                "   {column_name:<{width}} Model\n"
                "   {line}".format(
                    column_name="Name",
                    line="-" * (max_name_length + 16),
                    width=max_name_length + 1
                )
            ),
            choices=rows,
            default=default
        )


def select_resource_field(
    model: Type[models.Model],
    message: str = None,
    multiple: bool = False,
    predicate: Callable = None,
    prepend_choices: List[str] = None,
    append_choices: List[str] = None,
    default: Union[str, List[str]] = None
) -> Union[None, str, List[str]]:
    """
    Диалог выбора одного или нескольких полей, связанных с типом Resource.
    """
    choices_data = []
    for field in model._meta.get_fields(include_hidden=True):
        if not utils.is_resource_field(field):
            continue

        if predicate is not None and not predicate(field):
            continue

        choices_data.append(
            FieldChoice(
                name=field.name,
                field=field
            )
        )

    choices_data = sorted(
        choices_data,
        key=operator.attrgetter("name")
    )

    # Оформление опций в виде двух колонок.
    rows = []
    max_name_length = max((len(record.name) for record in choices_data), default=0)
    max_name_length = min(max(max_name_length, 10), 48)
    for choice in choices_data:
        rows.append([
            "{:<{width}} {}".format(
                choice.name,
                type(choice.field).__name__,
                width=max_name_length + 1
            ),
            choice.name
        ])

    if prepend_choices:
        rows = list(prepend_choices) + rows

    if append_choices:
        rows.extend(append_choices)

    if not rows:
        return

    if multiple:
        return prompt_variants(
            message=message or (
                "Select one or more resource fields:\n"
                "   {column_name:<{width}} Class\n"
                "   {line}".format(
                    column_name="Name",
                    line="-" * (max_name_length + 16),
                    width=max_name_length + 1
                )
            ),
            choices=rows,
            default=default
        )
    else:
        return prompt_action(
            message=message or (
                "Select resource field:\n"
                "   {column_name:<{width}} Class\n"
                "   {line}".format(
                    column_name="Name",
                    line="-" * (max_name_length + 16),
                    width=max_name_length + 1
                )
            ),
            choices=rows,
            default=default
        )


def select_collection_variations(
    collection: Type[CollectionBase],
    item_type: str,
    message: str = None,
    multiple: bool = False,
    predicate: Callable = None,
    prepend_choices: List[str] = None,
    append_choices: List[str] = None,
    default: Union[str, List[str]] = None
) -> Union[None, str, List[str]]:
    """
    Диалог выбора одной или нескольких вариаций элемента коллекции.
    """
    if item_type not in collection.item_types:
        raise RuntimeError("Unsupported collection item type: %s" % item_type)

    item_type_field = collection.item_types[item_type]
    if not utils.is_variations_allowed(item_type_field.model):
        raise RuntimeError("The specified collection item type does not support variations: %s" % item_type)

    choices_data = []
    for name, variation in utils.get_collection_variations(collection, item_type_field).items():
        if predicate is not None and not predicate(variation):
            continue

        choices_data.append(
            VariationChoice(
                name=name,
                size="×".join(map(str, variation.size))
            )
        )

    choices_data = sorted(
        choices_data,
        key=operator.attrgetter("name")
    )

    # Оформление опций в виде двух колонок.
    rows = []
    max_name_length = max((len(record.name) for record in choices_data), default=0)
    max_name_length = min(max(max_name_length, 10), 48)
    for choice in choices_data:
        rows.append([
            "{:<{width}} {}".format(
                choice.name,
                choice.size,
                width=max_name_length + 1
            ),
            choice.name
        ])

    rows = sorted(rows)

    if prepend_choices:
        rows = list(prepend_choices) + rows

    if append_choices:
        rows.extend(append_choices)

    if not rows:
        return

    if multiple:
        return prompt_variants(
            message=message or (
                "Select one or more variations:\n"
                "   {column_name:<{width}} Size\n"
                "   {line}".format(
                    column_name="Name",
                    line="-" * (max_name_length + 10),
                    width=max_name_length + 3
                )
            ),
            choices=rows,
            default=default
        )
    else:
        return prompt_action(
            message=message or (
                "Select variation:\n"
                "   {column_name:<{width}} Size\n"
                "   {line}".format(
                    column_name="Name",
                    line="-" * (max_name_length + 10),
                    width=max_name_length + 1
                )
            ),
            choices=rows,
            default=default
        )


def select_resource_variations(
    model: Type[models.Model],
    field_name: str,
    message: str = None,
    multiple: bool = False,
    predicate: Callable = None,
    prepend_choices: List[str] = None,
    append_choices: List[str] = None,
    default: Union[str, List[str]] = None
) -> Union[None, str, List[str]]:
    """
    Диалог выбора одной или нескольких вариаций изображения.
    """
    field = model._meta.get_field(field_name)

    if not utils.is_resource_field(field) or not utils.is_variations_allowed(field.related_model):
        raise RuntimeError("The specified field does not support variations: %s" % field.name)

    choices_data = []
    for name, variation in utils.get_field_variations(field).items():
        if predicate is not None and not predicate(variation):
            continue

        choices_data.append(
            VariationChoice(
                name=name,
                size="×".join(map(str, variation.size))
            )
        )

    choices_data = sorted(
        choices_data,
        key=operator.attrgetter("name")
    )

    # Оформление опций в виде двух колонок.
    rows = []
    max_name_length = max((len(record.name) for record in choices_data), default=0)
    max_name_length = min(max(max_name_length, 10), 48)
    for choice in choices_data:
        rows.append([
            "{:<{width}} {}".format(
                choice.name,
                choice.size,
                width=max_name_length + 1
            ),
            choice.name
        ])

    rows = sorted(rows)

    if prepend_choices:
        rows = list(prepend_choices) + rows

    if append_choices:
        rows.extend(append_choices)

    if not rows:
        return

    if multiple:
        return prompt_variants(
            message=message or (
                "Select one or more variations:\n"
                "   {column_name:<{width}} Size\n"
                "   {line}".format(
                    column_name="Name",
                    line="-" * (max_name_length + 10),
                    width=max_name_length + 3
                )
            ),
            choices=rows,
            default=default
        )
    else:
        return prompt_action(
            message=message or (
                "Select variation:\n"
                "   {column_name:<{width}} Size\n"
                "   {line}".format(
                    column_name="Name",
                    line="-" * (max_name_length + 10),
                    width=max_name_length + 1
                )
            ),
            choices=rows,
            default=default
        )



import hashlib
import re
from typing import Any, Dict, Iterable, Set, Tuple

from django.utils import formats
from django.utils.html import avoid_wrapping
from django.utils.translation import gettext, ngettext

from .typing import FileLike

filesize_regex = re.compile(r"^([.\d]+)\s*([KMGT])?B?$")
filesize_units = {"K": 10 ** 3, "M": 10 ** 6, "G": 10 ** 9, "T": 10 ** 12}


class cached_method:  # noqa: N801
    """
    Декоратор для кэширования результата вызова метода класса без параметров.
    Если декорируемый метод при выполнении вызовет исключение cached_method.Bypass
    с некоторым значением, то это значение будет возвращено как результат вызова метода,
    но не будет закэшировано.

    Пример:
        class Letters:
            def __init__(self, value: str = None):
                self.value = value

            @cached_method(key="_cache")
            def get_letters(self):
                if self.value is None:
                    raise cached_method.Bypass(set())

                return set(self.values)


        l = Letters()
        result = l.get_letters()
        assert result == set()    # пустое множество, которое не попало в кэш

        l.value = "paper"
        result = l.get_letters()
        assert result == {"p", "a", "e", "r"}  # множество было закэшировано
    """

    class Bypass(Exception):  # noqa: N818
        def __init__(self, value=None):
            self.value = value

    def __init__(self, key: str):
        self.key = key

    def __call__(self, func):
        def inner(instance):
            try:
                return getattr(instance, self.key)
            except AttributeError:
                pass

            try:
                value = func(instance)
            except self.Bypass as exc:
                return exc.value

            setattr(instance, self.key, value)
            return value

        # Привязка ключа кэша к функции-декоратору для возможности
        # очистки кэша извне метода.
        inner.cache_key = self.key

        return inner


def checksum(file: FileLike) -> str:
    """
    DropBox checksum realization.
    https://www.dropbox.com/developers/reference/content-hash
    """
    if file.closed:
        file.open("rb")
    elif file.seekable():
        file.seek(0)

    blocks = []
    while True:
        data = file.read(4 * 1024 * 1024)
        if not data:
            break
        blocks.append(hashlib.sha256(data).digest())
    return hashlib.sha256(b"".join(blocks)).hexdigest()


def remove_dulpicates(seq: Iterable) -> Tuple:
    """
    https://stackoverflow.com/questions/480214/how-do-you-remove-duplicates-from-a-list-whilst-preserving-order
    """
    seen = set()  # type: Set[Any]
    seen_add = seen.add
    return tuple(x for x in seq if not (x in seen or seen_add(x)))


def lowercased_dict_keys(options: Dict[str, Any]) -> Dict[str, Any]:
    """
    Возвращает копию словаря с ключами, приведенными к нижнему регистру.
    """
    return {key.lower(): value for key, value in options.items()}


def parse_filesize(value: str) -> int:
    """
    Парсинг человеко-понятного значения размера файла.
    Допустимые форматы значения: 4k, 4kb, 4KB, 4 K, 4 Kb
    """
    match = filesize_regex.match(value.upper().strip())
    number, unit = match.groups()
    return int(float(number) * filesize_units.get(unit, 1))


def filesizeformat(bytes_: int) -> str:
    """
    Форматирование размера файла.
    В отличие от одноименной встроенной функции, выводит настоящие килобайты/мегабайты,
    вместо кибибайтов/мибибайтов.
    """
    try:
        bytes_ = float(bytes_)
    except (TypeError, ValueError, UnicodeDecodeError):
        value = ngettext("%(size)d byte", "%(size)d bytes", 0) % {"size": 0}
        return avoid_wrapping(value)

    def filesize_number_format(value):
        return formats.number_format(round(value, 1), 1)

    KB = 10 ** 3  # noqa: N806
    MB = 10 ** 6  # noqa: N806
    GB = 10 ** 9  # noqa: N806
    TB = 10 ** 12  # noqa: N806
    PB = 10 ** 15  # noqa: N806

    negative = bytes_ < 0
    if negative:
        bytes_ = -bytes_  # Allow formatting of negative numbers.

    if bytes_ < KB:
        value = ngettext("%(size)d byte", "%(size)d bytes", bytes_) % {"size": bytes_}
    elif bytes_ < MB:
        value = gettext("%s KB") % filesize_number_format(bytes_ / KB)
    elif bytes_ < GB:
        value = gettext("%s MB") % filesize_number_format(bytes_ / MB)
    elif bytes_ < TB:
        value = gettext("%s GB") % filesize_number_format(bytes_ / GB)
    elif bytes_ < PB:
        value = gettext("%s TB") % filesize_number_format(bytes_ / TB)
    else:
        value = gettext("%s PB") % filesize_number_format(bytes_ / PB)

    if negative:
        value = "-%s" % value
    return avoid_wrapping(value)



import datetime
import posixpath


def generate_filename(instance, filename):
    # TODO: deprecated (migrations)
    upload_to = instance.get_file_folder()
    dirname = datetime.datetime.now().strftime(upload_to)
    return posixpath.join(dirname, filename)



from typing import IO, Any, Dict, List, Tuple, Union

from django.core.files import File

__all__ = ["FileLike", "Limitations", "VariationConfig"]

FileLike = Union[IO, File]
VariationConfig = Dict[str, Any]
Limitations = List[Union[str, Tuple[str, str]]]



# Generated by Django 4.0.4 on 2022-05-25 05:32

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('paper_uploads', '0008_svgitem_height_svgitem_width_uploadedsvgfile_height_and_more'),
    ]

    operations = [
        migrations.AddField(
            model_name='collection',
            name='concrete_collection_content_type',
            field=models.ForeignKey(editable=False, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='+', to='contenttypes.contenttype'),
        ),
        migrations.AddField(
            model_name='collectionitembase',
            name='concrete_collection_content_type',
            field=models.ForeignKey(editable=False, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='+', to='contenttypes.contenttype'),
        ),
        migrations.AlterField(
            model_name='collection',
            name='collection_content_type',
            field=models.ForeignKey(editable=False, on_delete=django.db.models.deletion.CASCADE, related_name='+', to='contenttypes.contenttype'),
        ),
        migrations.AlterField(
            model_name='collectionitembase',
            name='collection_content_type',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='contenttypes.contenttype'),
        ),
    ]



from django import forms

from ...models.collection import FileItem, ImageItem, SVGItem
from .base import ChangeFileResourceDialogBase


class ChangeFileItemDialog(ChangeFileResourceDialogBase):
    class Meta:
        model = FileItem
        fields = ["new_name", "display_name"]


class ChangeSVGItemDialog(ChangeFileResourceDialogBase):
    class Meta:
        model = SVGItem
        fields = ["new_name", "display_name", "title", "description"]
        widgets = {
            "description": forms.Textarea(attrs={
                "rows": 2
            })
        }


class ChangeImageItemDialog(ChangeFileResourceDialogBase):
    class Meta:
        model = ImageItem
        fields = ["new_name", "title", "description"]
        widgets = {
            "description": forms.Textarea(attrs={
                "rows": 2
            })
        }



from django.contrib import admin

from ..models.svg import UploadedSVGFile
from .file import UploadedFileAdmin


@admin.register(UploadedSVGFile)
class UploadedSVGFileAdmin(UploadedFileAdmin):
    pass



# Generated by Django 3.0.9 on 2020-09-15 07:08

import django.core.files.storage
from django.db import migrations, models
import django.db.models.deletion
import django.db.models.manager
import django.utils.timezone
import paper_uploads.models.fields.image
import paper_uploads.models.mixins


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
    ]

    operations = [
        migrations.CreateModel(
            name='Permissions',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
            ],
            options={
                'permissions': (('upload', 'Can upload files'), ('change', 'Can change files'), ('delete', 'Can delete files')),
                'managed': False,
                'default_permissions': (),
            },
        ),
        migrations.CreateModel(
            name='CollectionItemBase',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('collection_id', models.IntegerField()),
                ('item_type', models.CharField(db_index=True, editable=False, max_length=32, verbose_name='type')),
                ('order', models.IntegerField(default=0, editable=False, verbose_name='order')),
                ('collection_content_type', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='contenttypes.ContentType')),
                ('polymorphic_ctype', models.ForeignKey(editable=False, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='polymorphic_paper_uploads.collectionitembase_set+', to='contenttypes.ContentType')),
            ],
            options={
                'verbose_name': 'item',
                'verbose_name_plural': 'items',
                'default_permissions': (),
            },
        ),
        migrations.CreateModel(
            name='UploadedFile',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('owner_app_label', models.CharField(editable=False, max_length=100)),
                ('owner_model_name', models.CharField(editable=False, max_length=100)),
                ('owner_fieldname', models.CharField(editable=False, max_length=255)),
                ('created_at', models.DateTimeField(default=django.utils.timezone.now, editable=False, verbose_name='created at')),
                ('modified_at', models.DateTimeField(auto_now=True, verbose_name='changed at')),
                ('basename', models.CharField(editable=False, help_text='Human-readable resource name', max_length=255, verbose_name='basename')),
                ('extension', models.CharField(editable=False, help_text='Lowercase, without leading dot', max_length=32, verbose_name='extension')),
                ('size', models.PositiveIntegerField(default=0, editable=False, verbose_name='size')),
                ('checksum', models.CharField(editable=False, max_length=64, verbose_name='checksum')),
                ('uploaded_at', models.DateTimeField(default=django.utils.timezone.now, editable=False, verbose_name='uploaded at')),
                ('file', models.FileField(max_length=255, storage=django.core.files.storage.FileSystemStorage(), upload_to='files/%Y-%m-%d', verbose_name='file')),
                ('display_name', models.CharField(blank=True, max_length=255, verbose_name='display name')),
            ],
            options={
                'verbose_name': 'file',
                'verbose_name_plural': 'files',
                'abstract': False,
                'default_permissions': (),
            },
            bases=(paper_uploads.models.mixins.FileFieldProxyMixin, paper_uploads.models.mixins.FileProxyMixin, models.Model),
        ),
        migrations.CreateModel(
            name='UploadedImage',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('owner_app_label', models.CharField(editable=False, max_length=100)),
                ('owner_model_name', models.CharField(editable=False, max_length=100)),
                ('owner_fieldname', models.CharField(editable=False, max_length=255)),
                ('created_at', models.DateTimeField(default=django.utils.timezone.now, editable=False, verbose_name='created at')),
                ('modified_at', models.DateTimeField(auto_now=True, verbose_name='changed at')),
                ('basename', models.CharField(editable=False, help_text='Human-readable resource name', max_length=255, verbose_name='basename')),
                ('extension', models.CharField(editable=False, help_text='Lowercase, without leading dot', max_length=32, verbose_name='extension')),
                ('size', models.PositiveIntegerField(default=0, editable=False, verbose_name='size')),
                ('checksum', models.CharField(editable=False, max_length=64, verbose_name='checksum')),
                ('uploaded_at', models.DateTimeField(default=django.utils.timezone.now, editable=False, verbose_name='uploaded at')),
                ('title', models.CharField(blank=True, help_text='The title is being used as a tooltip when the user hovers the mouse over the image', max_length=255, verbose_name='title')),
                ('description', models.TextField(blank=True, help_text='This text will be used by screen readers, search engines, or when the image cannot be loaded', verbose_name='description')),
                ('width', models.PositiveSmallIntegerField(default=0, editable=False, verbose_name='width')),
                ('height', models.PositiveSmallIntegerField(default=0, editable=False, verbose_name='height')),
                ('cropregion', models.CharField(blank=True, editable=False, max_length=24, verbose_name='crop region')),
                ('file', paper_uploads.models.fields.image.VariationalFileField(max_length=255, storage=django.core.files.storage.FileSystemStorage(), upload_to='images/%Y-%m-%d', verbose_name='file')),
            ],
            options={
                'verbose_name': 'image',
                'verbose_name_plural': 'images',
                'abstract': False,
                'default_permissions': (),
            },
            bases=(paper_uploads.models.mixins.FileFieldProxyMixin, paper_uploads.models.mixins.FileProxyMixin, models.Model),
        ),
        migrations.CreateModel(
            name='FileItem',
            fields=[
                ('collectionitembase_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='paper_uploads.CollectionItemBase')),
                ('owner_app_label', models.CharField(editable=False, max_length=100)),
                ('owner_model_name', models.CharField(editable=False, max_length=100)),
                ('owner_fieldname', models.CharField(editable=False, max_length=255)),
                ('created_at', models.DateTimeField(default=django.utils.timezone.now, editable=False, verbose_name='created at')),
                ('modified_at', models.DateTimeField(auto_now=True, verbose_name='changed at')),
                ('basename', models.CharField(editable=False, help_text='Human-readable resource name', max_length=255, verbose_name='basename')),
                ('extension', models.CharField(editable=False, help_text='Lowercase, without leading dot', max_length=32, verbose_name='extension')),
                ('size', models.PositiveIntegerField(default=0, editable=False, verbose_name='size')),
                ('checksum', models.CharField(editable=False, max_length=64, verbose_name='checksum')),
                ('uploaded_at', models.DateTimeField(default=django.utils.timezone.now, editable=False, verbose_name='uploaded at')),
                ('file', models.FileField(max_length=255, storage=django.core.files.storage.FileSystemStorage(), upload_to='collections/files/%Y-%m-%d', verbose_name='file')),
                ('display_name', models.CharField(blank=True, max_length=255, verbose_name='display name')),
            ],
            options={
                'verbose_name': 'File item',
                'verbose_name_plural': 'File items',
                'abstract': False,
                'base_manager_name': 'objects',
            },
            bases=('paper_uploads.collectionitembase', paper_uploads.models.mixins.FileFieldProxyMixin, paper_uploads.models.mixins.FileProxyMixin, models.Model),
        ),
        migrations.CreateModel(
            name='ImageItem',
            fields=[
                ('collectionitembase_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='paper_uploads.CollectionItemBase')),
                ('owner_app_label', models.CharField(editable=False, max_length=100)),
                ('owner_model_name', models.CharField(editable=False, max_length=100)),
                ('owner_fieldname', models.CharField(editable=False, max_length=255)),
                ('created_at', models.DateTimeField(default=django.utils.timezone.now, editable=False, verbose_name='created at')),
                ('modified_at', models.DateTimeField(auto_now=True, verbose_name='changed at')),
                ('basename', models.CharField(editable=False, help_text='Human-readable resource name', max_length=255, verbose_name='basename')),
                ('extension', models.CharField(editable=False, help_text='Lowercase, without leading dot', max_length=32, verbose_name='extension')),
                ('size', models.PositiveIntegerField(default=0, editable=False, verbose_name='size')),
                ('checksum', models.CharField(editable=False, max_length=64, verbose_name='checksum')),
                ('uploaded_at', models.DateTimeField(default=django.utils.timezone.now, editable=False, verbose_name='uploaded at')),
                ('title', models.CharField(blank=True, help_text='The title is being used as a tooltip when the user hovers the mouse over the image', max_length=255, verbose_name='title')),
                ('description', models.TextField(blank=True, help_text='This text will be used by screen readers, search engines, or when the image cannot be loaded', verbose_name='description')),
                ('width', models.PositiveSmallIntegerField(default=0, editable=False, verbose_name='width')),
                ('height', models.PositiveSmallIntegerField(default=0, editable=False, verbose_name='height')),
                ('cropregion', models.CharField(blank=True, editable=False, max_length=24, verbose_name='crop region')),
                ('file', paper_uploads.models.fields.image.VariationalFileField(max_length=255, storage=django.core.files.storage.FileSystemStorage(), upload_to='collections/images/%Y-%m-%d', verbose_name='file')),
            ],
            options={
                'verbose_name': 'Image item',
                'verbose_name_plural': 'Image items',
                'abstract': False,
                'base_manager_name': 'objects',
            },
            bases=('paper_uploads.collectionitembase', paper_uploads.models.mixins.FileFieldProxyMixin, paper_uploads.models.mixins.FileProxyMixin, models.Model),
        ),
        migrations.CreateModel(
            name='SVGItem',
            fields=[
                ('collectionitembase_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='paper_uploads.CollectionItemBase')),
                ('owner_app_label', models.CharField(editable=False, max_length=100)),
                ('owner_model_name', models.CharField(editable=False, max_length=100)),
                ('owner_fieldname', models.CharField(editable=False, max_length=255)),
                ('created_at', models.DateTimeField(default=django.utils.timezone.now, editable=False, verbose_name='created at')),
                ('modified_at', models.DateTimeField(auto_now=True, verbose_name='changed at')),
                ('basename', models.CharField(editable=False, help_text='Human-readable resource name', max_length=255, verbose_name='basename')),
                ('extension', models.CharField(editable=False, help_text='Lowercase, without leading dot', max_length=32, verbose_name='extension')),
                ('size', models.PositiveIntegerField(default=0, editable=False, verbose_name='size')),
                ('checksum', models.CharField(editable=False, max_length=64, verbose_name='checksum')),
                ('uploaded_at', models.DateTimeField(default=django.utils.timezone.now, editable=False, verbose_name='uploaded at')),
                ('file', models.FileField(max_length=255, storage=django.core.files.storage.FileSystemStorage(), upload_to='collections/files/%Y-%m-%d', verbose_name='file')),
                ('display_name', models.CharField(blank=True, max_length=255, verbose_name='display name')),
            ],
            options={
                'verbose_name': 'SVG item',
                'verbose_name_plural': 'SVG items',
                'abstract': False,
                'base_manager_name': 'objects',
            },
            bases=('paper_uploads.collectionitembase', paper_uploads.models.mixins.FileFieldProxyMixin, paper_uploads.models.mixins.FileProxyMixin, models.Model),
        ),
        migrations.CreateModel(
            name='Collection',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('owner_app_label', models.CharField(editable=False, max_length=100)),
                ('owner_model_name', models.CharField(editable=False, max_length=100)),
                ('owner_fieldname', models.CharField(editable=False, max_length=255)),
                ('created_at', models.DateTimeField(default=django.utils.timezone.now, editable=False, verbose_name='created at')),
                ('collection_content_type', models.ForeignKey(editable=False, null=True, on_delete=django.db.models.deletion.SET_NULL, to='contenttypes.ContentType')),
            ],
            options={
                'proxy': False,
                'default_manager_name': 'default_mgr',
            },
            managers=[
                ('default_mgr', django.db.models.manager.Manager()),
            ],
        ),
        migrations.CreateModel(
            name='ImageCollection',
            fields=[
            ],
            options={
                'proxy': True,
                'indexes': [],
            },
            bases=('paper_uploads.collection',),
            managers=[
                ('default_mgr', django.db.models.manager.Manager()),
            ],
        ),
    ]



import os
from typing import Any, Generator, Tuple, Type, Union, cast

from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ObjectDoesNotExist
from django.core.files.uploadedfile import UploadedFile
from django.core.handlers.wsgi import WSGIRequest
from django.db import transaction
from django.http import HttpResponse
from django.utils.decorators import method_decorator
from django.utils.module_loading import import_string
from django.utils.translation import gettext_lazy as _
from django.views.decorators.csrf import csrf_exempt

from .. import exceptions, signals
from ..helpers import run_validators
from ..logging import logger
from ..models.collection import CollectionBase, CollectionFileItemBase, CollectionItemBase
from ..models.fields.collection import CollectionItem
from ..models.mixins import EditableResourceMixin
from . import helpers
from .base import AjaxView, ChangeFileViewBase, DeleteFileViewBase, UploadFileViewBase


class CreateCollectionView(AjaxView):
    http_method_names = ["post"]

    @method_decorator(csrf_exempt)
    def dispatch(self, *args, **kwargs):
        return super().dispatch(*args, **kwargs)

    def post(self, request: WSGIRequest, *args, **kwargs) -> HttpResponse:
        if not request.user.has_perm("paper_uploads.upload"):
            return self.error_response(_("Access denied"))
        return self.wrap(self.handle)()

    def get_collection_model(self) -> Type[CollectionBase]:
        content_type_id = self.request.POST.get("paperCollectionContentType")
        return helpers.get_model_class(content_type_id, CollectionBase)

    def get_instance(self) -> CollectionBase:
        collection_cls = self.get_collection_model()
        return collection_cls(
            owner_app_label=self.request.POST.get("paperOwnerAppLabel"),
            owner_model_name=self.request.POST.get("paperOwnerModelName"),
            owner_fieldname=self.request.POST.get("paperOwnerFieldName"),
        )

    def handle(self) -> HttpResponse:
        instance = self.get_instance()
        instance.save()
        return self.success(instance)

    def success(self, instance: CollectionBase) -> HttpResponse:
        return self.success_response({
            "collection_id": instance.pk
        })


class DeleteCollectionView(AjaxView):
    http_method_names = ["post"]

    @method_decorator(csrf_exempt)
    def dispatch(self, *args, **kwargs):
        return super().dispatch(*args, **kwargs)

    def post(self, request: WSGIRequest, *args, **kwargs) -> HttpResponse:
        if not request.user.has_perm("paper_uploads.delete"):
            return self.error_response(_("Access denied"))
        return self.wrap(self.handle)()

    def get_collection_model(self) -> Type[CollectionBase]:
        content_type_id = self.request.POST.get("paperCollectionContentType")
        return helpers.get_model_class(content_type_id, CollectionBase)

    def get_collection_id(self) -> Any:
        return self.request.POST.get("collectionId")

    def get_instance(self) -> CollectionBase:
        collection_cls = self.get_collection_model()
        collection_id = self.get_collection_id()
        return cast(CollectionBase, helpers.get_instance(collection_cls, collection_id))

    def handle(self) -> HttpResponse:
        collection = self.get_instance()
        collection.delete()
        return self.success()

    def success(self) -> HttpResponse:
        return self.success_response()


class UploadFileView(UploadFileViewBase):
    def get_collection_model(self) -> Type[CollectionBase]:
        content_type_id = self.request.POST.get("paperCollectionContentType")
        return helpers.get_model_class(content_type_id, CollectionBase)

    def get_collection_id(self) -> Any:
        return self.request.POST.get("collectionId")

    def get_collection_instance(self) -> CollectionBase:
        collection_cls = self.get_collection_model()
        collection_id = self.get_collection_id()
        return cast(CollectionBase, helpers.get_instance(collection_cls, collection_id))

    def get_instance(self, item_type: str, **kwargs) -> CollectionItemBase:
        collection_cls = self.get_collection_model()
        item_model = collection_cls.get_item_model(item_type)
        return item_model(
            type=item_type,
            **kwargs
        )

    def get_order(self) -> int:
        order = self.request.POST.get("order")

        try:
            order = max(0, int(order))
        except (TypeError, ValueError):
            order = None

        return order

    def get_accepted_item_types(
        self,
        collection: Union[CollectionBase, Type[CollectionBase]],
        file: UploadedFile
    ) -> Generator[Tuple[str, CollectionItem], Any, None]:
        for item_type, field in collection.item_types.items():
            if issubclass(field.model, CollectionFileItemBase) and field.model.accept(file):
                yield item_type, field

    def handle(self, file: UploadedFile) -> HttpResponse:
        collection = self.get_collection_instance()

        # перебираем все подходящие классы элементов пока
        # не найдем тот, который будет успешно создан
        for item_type, item_type_field in self.get_accepted_item_types(collection, file):
            item = self.get_instance(
                collection=collection,
                item_type=item_type,
                size=file.size,
                order=self.get_order()
            )

            try:
                item.attach(file)
            except exceptions.UnsupportedResource:
                continue

            try:
                item.full_clean()
                run_validators(file, item_type_field.validators)
            except Exception:
                item.delete_file()
                raise

            break
        else:
            filename = os.path.basename(file.name)
            return self.error_response(_("Unsupported file: %s") % filename)

        item.save()

        if not file.closed:
            file.close()

        return self.success(item)

    def success(self, instance: CollectionItemBase) -> HttpResponse:
        return self.success_response(instance.as_dict())


class DeleteFileView(DeleteFileViewBase):
    def get_collection_model(self) -> Type[CollectionBase]:
        content_type_id = self.request.POST.get("paperCollectionContentType")
        return helpers.get_model_class(content_type_id, CollectionBase)

    def get_item_type(self) -> str:
        return self.request.POST.get("itemType")

    def get_item_id(self) -> Any:
        return self.request.POST.get("itemId")

    def get_instance(self) -> CollectionItemBase:
        collection_cls = self.get_collection_model()
        item_type = self.get_item_type()
        item_model = collection_cls.get_item_model(item_type)
        item_id = self.get_item_id()
        return cast(CollectionItemBase, helpers.get_instance(item_model, item_id))

    def handle(self) -> HttpResponse:
        try:
            item = self.get_instance()
        except ObjectDoesNotExist:
            # silently skip
            pass
        else:
            item.delete()

        return self.success()

    def success(self) -> HttpResponse:
        return self.success_response()


class ChangeFileView(ChangeFileViewBase):
    template_name = "paper_uploads/dialogs/collection.html"

    def get_collection_model(self) -> Type[CollectionBase]:
        content_type_id = self.request.GET.get("paperCollectionContentType")
        return helpers.get_model_class(content_type_id, CollectionBase)

    def get_item_type(self) -> str:
        return self.request.GET.get("itemType")

    def get_item_id(self) -> Any:
        return self.request.GET.get("itemId")

    def get_instance(self) -> CollectionItemBase:
        collection_cls = self.get_collection_model()
        item_id = self.get_item_id()
        item_type = self.get_item_type()
        item_model = collection_cls.get_item_model(item_type)
        return cast(CollectionItemBase, helpers.get_instance(item_model, item_id))

    def get_form_class(self):
        if self.form_class is not None:
            return self.form_class

        collection_cls = self.get_collection_model()
        item_type = self.get_item_type()
        item_model = collection_cls.get_item_model(item_type)
        if issubclass(item_model, EditableResourceMixin):
            return import_string(item_model.change_form_class)


class SortItemsView(AjaxView):
    http_method_names = ["post"]

    @method_decorator(csrf_exempt)
    def dispatch(self, *args, **kwargs):
        return super().dispatch(*args, **kwargs)

    def post(self, request: WSGIRequest, *args, **kwargs) -> HttpResponse:
        if not request.user.has_perm("paper_uploads.change"):
            return self.error_response(_("Access denied"))
        return self.wrap(self.handle)()

    def get_collection_model(self) -> Type[CollectionBase]:
        content_type_id = self.request.POST.get("paperCollectionContentType")
        return helpers.get_model_class(content_type_id, CollectionBase)

    def get_collection_id(self) -> Any:
        return self.request.POST.get("collectionId")

    def handle(self) -> HttpResponse:
        collection_cls = self.get_collection_model()
        collection_id = self.get_collection_id()
        collection_content_type = ContentType.objects.get_for_model(collection_cls, for_concrete_model=False)
        collection = helpers.get_instance(collection_cls, collection_id)

        # Получение списка ID элементов в том порядке, в котором они должны быть расположены
        order_string = self.request.POST.get("orderList", "")
        ordered_item_ids = (item.strip() for item in order_string.split(","))
        ordered_item_ids = tuple(pk for pk in ordered_item_ids if pk)

        # Список ID и сортировки для существующих элементов коллекции
        existing_item_ids = {
            str(pk): order
            for pk, order in CollectionItemBase._base_manager.filter(
                collection_content_type=collection_content_type,
                collection_id=collection_id
            ).values_list("pk", "order")
        }

        with transaction.atomic():
            for index, item_id in enumerate(ordered_item_ids):
                if item_id in existing_item_ids:
                    # обновляем только те значения сортировки, которые изменились
                    if existing_item_ids[item_id] != index:
                        CollectionItemBase._base_manager.filter(
                            pk=item_id
                        ).update(
                            order=index
                        )
                else:
                    logger.warning(
                        "Item #{} not found in collection #{}".format(item_id, collection_id)
                    )

        signals.collection_reordered.send(
            sender=collection_cls,
            instance=collection
        )

        return self.success_response({
            "orderMap": dict(CollectionItemBase._base_manager.filter(
                collection_content_type=collection_content_type,
                collection_id=collection_id
            ).values_list("pk", "order"))
        })



# Generated by Django 3.2.9 on 2021-12-02 18:59

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('paper_uploads', '0006_auto_20211116_0748'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='collection',
            options={'default_permissions': ()},
        ),
    ]



from django.apps import AppConfig
from django.db.models.signals import pre_migrate
from django.utils.translation import gettext_lazy as _


class Config(AppConfig):
    name = "paper_uploads"
    verbose_name = _("Uploaded Files")

    def ready(self):
        from .signals import handlers

        pre_migrate.connect(handlers.inject_operations, sender=self)



import sys
from enum import Enum, auto

from django.apps import apps
from django.core.management import BaseCommand
from django.db import DEFAULT_DB_ALIAS

from paper_uploads.management import helpers

from .. import utils


class Step(Enum):
    GET_MODEL = auto()
    GET_FIELD = auto()
    GET_VARIATIONS = auto()
    PROCESS = auto()
    END = auto()


class ExitException(Exception):
    pass


class Command(BaseCommand):
    help = """
    Удаление вариаций для всех экземпляров указанной модели.

    Пример для обычной модели:
        python3 manage.py remove_variations blog.post --field=hero

    Пример для коллекции:
        python3 manage.py remove_variations blog.gallery --item-type=image
    """
    options = None
    verbosity = None
    database = DEFAULT_DB_ALIAS
    requires_system_checks = []

    _step = Step.GET_MODEL
    _model = None
    _is_collection = None
    _field_name = None
    _variation_names = None

    def add_arguments(self, parser):
        parser.add_argument(
            "--model",
            metavar="[APP_LABEL].[MODEL_NAME]",
            help="Specifies the model to remove variation files for",
        )
        parser.add_argument(
            "--field",
            nargs="?",
            help="Restricts variations to the specified field. "
                 "You should not specify any field for Collection models.",
        )
        parser.add_argument(
            "--item-type",
            nargs="?",
            help="Only look for variations in the specified CollectionItem. "
                 "Use this argument for Collection models only.",
        )
        parser.add_argument(
            "--variations",
            dest="variations",
            nargs="+",
            help="Specifies the variation names to delete",
        )
        parser.add_argument(
            "--database",
            action="store",
            dest="database",
            default=DEFAULT_DB_ALIAS,
            help="Nominates the database to use. Defaults to the 'default' database.",
        )

    def handle(self, *args, **options):
        self.options = options
        self.verbosity = options["verbosity"]
        self.database = options["database"]

        try:
            self.loop()
        except ExitException:
            return

    def loop(self):
        while True:
            if self._step is Step.GET_MODEL:
                self.get_model()
            elif self._step is Step.GET_FIELD:
                if self._is_collection:
                    self.get_collection_field()
                else:
                    self.get_resource_field()
            elif self._step is Step.GET_VARIATIONS:
                if self._is_collection:
                    self.get_collection_variations()
                else:
                    self.get_resource_variations()
            elif self._step is Step.PROCESS:
                if self._is_collection:
                    self.process_collection()
                else:
                    self.process_resource()
            else:
                return

    def get_model(self):
        model_name = self.options["model"]
        if model_name is None:
            model_name = helpers.select_resource_model(
                append_choices=["[Exit]"],
                predicate=lambda model: utils.includes_variations(model),
            )

        if model_name == "[Exit]":
            raise ExitException

        self._model = apps.get_model(model_name)
        self._is_collection = utils.is_collection(self._model)
        self._step = Step.GET_FIELD

    def get_collection_field(self):
        item_type = self.options["item_type"]
        if item_type is None:
            item_type = helpers.select_collection_item_type(
                self._model,
                predicate=lambda f: utils.is_variations_allowed(f.model),
                append_choices=["[Back]", "[Exit]"]
            )

        if item_type == "[Exit]":
            raise ExitException

        if item_type == "[Back]":
            self._step = Step.GET_MODEL
            return

        self._field_name = item_type
        self._step = Step.GET_VARIATIONS

    def get_resource_field(self):
        field_name = self.options["field"]
        if field_name is None:
            field_name = helpers.select_resource_field(
                self._model,
                predicate=lambda f: utils.is_variations_allowed(f.related_model),
                append_choices=["[Back]", "[Exit]"]
            )

        if field_name == "[Exit]":
            raise ExitException

        if field_name == "[Back]":
            self._step = Step.GET_MODEL
            return

        self._field_name = field_name
        self._step = Step.GET_VARIATIONS

    def get_collection_variations(self):
        variations = self.options["variations"]
        if variations is None:
            variations = helpers.select_collection_variations(
                self._model,
                self._field_name,
                multiple=True,
                prepend_choices=["[All]"],
                append_choices=["[Back]", "[Exit]"],
                default="[All]"
            )

        if "[Exit]" in variations:
            raise ExitException

        if "[Back]" in variations:
            self._step = Step.GET_FIELD
            return

        if "[All]" in variations:
            item_type_field = self._model.item_types[self._field_name]
            self._variation_names = set(utils.get_collection_variations(
                self._model,
                item_type_field
            ))
        else:
            self._variation_names = set(variations)

        self._step = Step.PROCESS

    def get_resource_variations(self):
        variations = self.options["variations"]
        if variations is None:
            variations = helpers.select_resource_variations(
                self._model,
                self._field_name,
                multiple=True,
                prepend_choices=["[All]"],
                append_choices=["[Back]", "[Exit]"],
                default="[All]"
            )

        if "[Exit]" in variations:
            raise ExitException

        if "[Back]" in variations:
            self._step = Step.GET_FIELD
            return

        if "[All]" in variations:
            field = self._model._meta.get_field(self._field_name)
            self._variation_names = set(utils.get_field_variations(field))
        else:
            self._variation_names = set(variations)

        self._step = Step.PROCESS

    def process_collection(self):
        if not self._variation_names:
            return

        queryset = self._model.objects.using(self.database)

        total = queryset.count()
        for index, collection in enumerate(queryset.iterator(), start=1):
            collection_items = collection.get_items(self._field_name)
            collection_item_count = collection_items.count()
            if not collection_item_count:
                continue

            print(
                "Processing \033[92m{}\033[0m items"
                " of \033[92m'{}.{}'\033[0m (ID: {}) ({}/{}) ... ".format(
                    collection_item_count,
                    self._model._meta.app_label,
                    self._model.__name__,
                    collection.pk,
                    index,
                    total
                ),
                end=""
            )
            sys.stdout.flush()

            for item in collection_items.iterator():
                for variation_name in item.get_variations():
                    if variation_name in self._variation_names:
                        variation_file = item.get_variation_file(variation_name)
                        variation_file.delete()

            print("done")
            sys.stdout.flush()

        self._step = Step.END

    def process_resource(self):
        if not self._variation_names:
            return

        queryset = self._model.objects.using(self.database).exclude((self._field_name, None))

        total = queryset.count()
        for index, instance in enumerate(queryset.iterator(), start=1):
            field = getattr(instance, self._field_name)
            if not field.file_exists():
                continue

            print(
                "Processing \033[92m'{}.{}'\033[0m (ID: {}) ({}/{}) ... ".format(
                    type(instance)._meta.app_label,
                    type(instance).__name__,
                    instance.pk,
                    index,
                    total
                ),
                end=""
            )
            sys.stdout.flush()

            for variation_name in field.get_variations():
                if variation_name in self._variation_names:
                    variation_file = field.get_variation_file(variation_name)
                    variation_file.delete()

            print("done")
            sys.stdout.flush()

        self._step = Step.END



import posixpath

from variations.variation import Variation


class PaperVariation(Variation):
    """
    Расширение возможностей вариации:
      * Хранение имени вариации
    """

    def __init__(self, *args, name: str = "", **kwargs):
        self.name = name
        super().__init__(*args, **kwargs)

    @property
    def name(self) -> str:
        return self._name

    @name.setter
    def name(self, value: str):
        if not isinstance(value, str):
            raise TypeError(value)
        self._name = value

    def get_output_filename(self, input_filename: str) -> str:
        """
        Конструирует имя файла для вариации по имени файла исходника.
        Имя файла может включать путь — он остается неизменным.
        """
        if not self.name:
            raise RuntimeError("`name` is empty")

        dir_name, file_name = posixpath.split(input_filename)
        file_root, file_ext = posixpath.splitext(file_name)
        new_file_root = posixpath.extsep.join((file_root, self.name))
        file_name = "".join((new_file_root, file_ext))
        name = posixpath.join(dir_name, file_name)
        return self.replace_extension(name)



from django.utils.functional import LazyObject

from .conf import settings
from .utils import lowercased_dict_keys


class UploadStorage(LazyObject):
    def _setup(self):
        options = lowercased_dict_keys(settings.STORAGE_OPTIONS)
        self._wrapped = settings.STORAGE(**options)


default_storage = UploadStorage()
upload_storage = default_storage  # deprecated
